#include <sys/stat.h>
#include <sys/types.h>
#ifndef _WIN32
#include <dirent.h>
#endif
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef _WIN32
#include <windows.h>
#endif

#ifndef MAX_PATH
#define MAX_PATH 4096
#endif

#ifdef _MSC_VER
#define snprintf _snprintf
#endif

typedef struct {
  int filename_length;
  char *filename;
  int offset;
  int length;
} tbf_t;

char filename[MAX_PATH + 4];

int sort_function(const void *a, const void *b) {
  tbf_t *one, *two;
 
  if (! a || ! b) return 0;

  one = *(tbf_t **) a;
  two = *(tbf_t **) b;

  return strcmp(one->filename, two->filename);
}

void usage() {
  printf("tbfpack: Crude .tbf file packer.\n");
  printf("usage: tbfpack directory [file.tbf]\n");
  exit(1);
}

void cleanup_records(tbf_t **records, int num_records) {
  tbf_t *record;
  int i;

  for (i = 0; i < num_records; i++) {
    record = records[i];
    if (record->filename) free(record->filename);
    free(record);
  }

  free(records);
}

/* dirname + '/' + basename => filename */
int make_path(char *filename, char *dirname, char *basename) {
#ifdef _WIN32
  if (snprintf(filename, MAX_PATH, "%s\\%s", dirname, basename) < 0) {
#else
  if (snprintf(filename, MAX_PATH, "%s/%s", dirname, basename) < 0) {
#endif
    fprintf(stderr, "Error making path %s/%s: %s\n", dirname, basename, strerror(errno));
    return 1;
  }
  return 0;
}

/* Check it's a file we can read */
int is_directory(char *dirname, char *basename) {
  struct stat st;
  int fd;

  if (make_path(filename, dirname, basename)) exit(111);
#ifdef _WIN32
  return GetFileAttributes(filename) & FILE_ATTRIBUTE_DIRECTORY;
#else
  if (stat(filename, &st)) {
    fprintf(stderr, "Can't stat '%s': %s\n", filename, strerror(errno));
    exit(112);
  }

  if (S_ISDIR(st.st_mode)) return 1;
  return 0;
#endif
}

tbf_t *alloc_tbf() {
  return (tbf_t *) malloc(sizeof(tbf_t));
}

tbf_t **alloc_tbfs(int num_tbfs) {
  int i, j;
  tbf_t **ret = (tbf_t **) malloc(sizeof(tbf_t *) * num_tbfs);
  if (! ret) return 0;

  for (i = 0; i < num_tbfs; i++) {
    ret[i] = alloc_tbf();
    if (! ret[i]) {
      for (j = 0; j < i; j++) {
        free(ret[j]);
        free(ret);
        return 0;
      }
    }
  }

  return ret;
}

int main(int argc, char **argv) {
  int num_files, i, offset;
  int fd, in, out, outfd, need, wrote;
#ifdef _WIN32
  HANDLE handle;
  WIN32_FIND_DATA info;
#else
  DIR *directory;
  struct dirent *d;
#endif
  tbf_t **records, *record;
  char buffer[4096];
  char *s;

  if (argc == 1) usage();

  /* Find number of files in directory */
  num_files = 0;
  offset = 4;
#ifdef _WIN32
  snprintf(filename, sizeof(filename), "%s\\*", argv[1]);
  handle = FindFirstFile(filename, &info);
  if (handle == INVALID_HANDLE_VALUE) {
    fprintf(stderr, "Error reading directory '%s': error %d\n", argv[1], GetLastError());
    exit(1);
  }

  do {
    /* Check it isn't a directory */
    if (info.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) /* nop */;
    else num_files++;

    /* Get next file */
    fd = FindNextFile(handle, &info);
  } while (fd);

  if (GetLastError() != ERROR_NO_MORE_FILES) {
    fprintf(stderr, "Error reading directory '%s': error %d\n", argv[1], GetLastError());
    FindClose(handle);
    exit(1);
  }

  FindClose(handle);
#else
  directory = opendir(argv[1]);
  if (! directory) {
    fprintf(stderr, "Can't open directory '%s': %s\n", argv[1], strerror(errno));
    exit(1);
  }

  while (d = readdir(directory)) {
    if (! strncmp(d->d_name, ".", 2)) continue;
    if (! strncmp(d->d_name, "..", 3)) continue;
    if (is_directory(argv[1], d->d_name)) continue;
    num_files++;
  }
#endif

  /* Allocate enough structs */
  records = alloc_tbfs(num_files);
  if (! records) {
    fprintf(stderr, "Out of memory for %d records\n: %s\n", num_files, strerror(errno));
    exit(2);
  }

  /* Now fill in our structs */
#ifdef _WIN32
  i = 0;

  handle = FindFirstFile(filename, &info);
  if (handle == INVALID_HANDLE_VALUE) {
    fprintf(stderr, "Error reading directory '%s': error %d\n", argv[1], GetLastError());
    exit(1);
  }

  do {
    /* Check it isn't a directory */
    if (info.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) /* nop */;
    else {
      /* Set the filename */
      record = records[i];
      record->filename_length = strlen(info.cFileName);
      offset += 12 + record->filename_length;
      record->filename = (char *) malloc(record->filename_length + 1);
      if (! record->filename) {
        fprintf(stderr, "Out of memory for record #%d ('%s'): error %d\n", i, info.cFileName, GetLastError());
        FindClose(handle);
        cleanup_records(records, num_files);
        exit(2);
      }
      memmove(record->filename, info.cFileName, record->filename_length);
      record->filename[record->filename_length] = '\0';

      /* Set the length */
      if (info.nFileSizeHigh) {
        fprintf(stderr, "File #%d ('%s') is too large!\n", i, info.cFileName);
        FindClose(handle);
        cleanup_records(records, num_files);
        exit(3);
      }
      record->length = info.nFileSizeLow;

      /* Next */
      i++;
    }

    /* Get next file */
    fd = FindNextFile(handle, &info);
  } while (fd);

  if (GetLastError() != ERROR_NO_MORE_FILES) {
    fprintf(stderr, "Error reading directory '%s': error %d\n", argv[1], GetLastError());
    FindClose(handle);
    exit(1);
  }

  FindClose(handle);
#else
  seekdir(directory, 0);
  for (i = 0; i < num_files; i++) {
    d = readdir(directory);
    if (! strncmp(d->d_name, ".", 2)) d = readdir(directory);
    if (! strncmp(d->d_name, "..", 3)) d = readdir(directory);
    if (is_directory(argv[1], d->d_name)) d = readdir(directory);
    record = records[i];

    /* Set filename */
    record->filename_length = strlen(d->d_name);
    offset += 12 + record->filename_length;
    record->filename = (char *) malloc(record->filename_length + 1);
    if (! record->filename) {
      fprintf(stderr, "Out of memory for record #%d ('%s'): %s\n", i, d->d_name, strerror(errno));
      closedir(directory);
      cleanup_records(records, num_files);
      exit(3);
    }
    memmove(record->filename, d->d_name, record->filename_length);
    record->filename[record->filename_length] = '\0';

    /* Construct the real path to the file we'll be archiving */
    if (make_path(filename, argv[1], record->filename)) {
      closedir(directory);
      cleanup_records(records, num_files);
      exit(4);
    }

    /* Open file and see how big it is */
    fd = open(filename, O_RDONLY);
    if (fd < 0) {
      fprintf(stderr, "Can't open file #%d ('%s') for reading: %s\n", i, filename, strerror(errno));
      closedir(directory);
      cleanup_records(records, num_files);
      exit(5);
    }

    /* Set offset and length */
    record->length = lseek(fd, 0, SEEK_END);
    if (record->length < 0) {
      fprintf(stderr, "Can't determine length of file #%d ('%s'): %s\n", i, filename, strerror(errno));
      closedir(directory);
      cleanup_records(records, num_files);
      exit(6);
    }

    /* Cleanup that file descriptor */
    close(fd);
  }

  /* Close the directory */
  closedir(directory);
#endif

  /* Sort the records */
  qsort((void *) records, num_files, sizeof(tbf_t *), sort_function);

  /* Calculate offsets */
  records[0]->offset = offset;
  for (i = 1; i < num_files; i++) {
    records[i]->offset = records[i - 1]->offset + records[i - 1]->length;
  }

  /* Get the filename */
  if (argc > 2) snprintf(filename, sizeof(filename), "%s", argv[2]);
  else {
    snprintf(filename, sizeof(filename), "%s", argv[1]);
    for (s = filename; *s; s++); s--;
#ifdef _WIN32
    for ( ; *s == '\\'; s--); s++;
#else
    for ( ; *s == '/'; s--); s++;
#endif
    snprintf(s, 5, "%s", ".tbf");
  }

  /* Open the file */
#ifdef _WIN32
  outfd = open(filename, O_BINARY | O_WRONLY | O_TRUNC | O_CREAT, 0644);
#else
  outfd = open(filename, O_WRONLY | O_TRUNC | O_CREAT, 0644);
#endif
  if (outfd < 0) {
    fprintf(stderr, "Error opening file '%s' for writing: %s\n", filename, strerror(errno));
    cleanup_records(records, num_files);
    exit(7);
  }

  /* Write the number of files */
  out = write(outfd, &num_files, 4);
  if (out != 4) {
    fprintf(stderr, "Error writing number of files: %s\n", strerror(errno));
    cleanup_records(records, num_files);
    close(outfd);
    exit(8);
  }

  /* Write details for each file */
  for (i = 0; i < num_files; i++) {
    record = records[i];

    /* Name length */
    out = write(outfd, &record->filename_length, 4);
    if (out != 4) {
      fprintf(stderr, "Error writing file #%d name length: %s\n", i, strerror(errno));
      cleanup_records(records, num_files);
      close(outfd);
      exit(9);
    }

    /* Name */
    out = write(outfd, record->filename, record->filename_length);
    if (out != record->filename_length) {
      fprintf(stderr, "Error writing file #%d name: %s\n", i, strerror(errno));
      cleanup_records(records, num_files);
      close(outfd);
      exit(10);
    }

    /* Offset */
    out = write(outfd, &record->offset, 4);
    if (out != 4) {
      fprintf(stderr, "Error writing file #%d offset: %s\n", i, strerror(errno));
      cleanup_records(records, num_files);
      close(outfd);
      exit(11);
    }

    /* Length */
    out = write(outfd, &record->length, 4);
    if (out != 4) {
      fprintf(stderr, "Error writing file #%d length: %s\n", i, strerror(errno));
      cleanup_records(records, num_files);
      close(outfd);
      exit(12);
    }
  }

  /* Process each file */
  for (i = 0; i < num_files; i++) {
    record = records[i];
    if (make_path(filename, argv[1], record->filename)) {
      cleanup_records(records, num_files);
      close(outfd);
      exit(13);
    }

    printf("Packing '%s'\n", record->filename);

    /* Open the file */
#ifdef _WIN32
    fd = open(filename, O_BINARY | O_RDONLY);
#else
    fd = open(filename, O_RDONLY);
#endif
    if (fd < 0) {
      fprintf(stderr, "Can't open file '%s' for reading: %s\n", filename, strerror(errno));
      cleanup_records(records, num_files);
      close(fd);
      close(outfd);
      exit(14);
    }

    /* Read it! */
    wrote = 0;
    while (wrote < record->length) {
      /* What we need to write */
      need = record->length - wrote;
      if (need > sizeof(buffer)) need = sizeof(buffer);

      /* Read a bit */
      in = read(fd, &buffer, need);
      if (in != need) {
        fprintf(stderr, "Error reading from position %d of file '%s': %s\n", wrote, filename, strerror(errno));
        cleanup_records(records, num_files);
        close(fd);
        close(outfd);
        exit(15);
      }

      /* Write a bit */
      out = write(outfd, &buffer, in);
      if (out != in) {
        fprintf(stderr, "Error writing from position %d of file '%s': %s\n", wrote, filename, strerror(errno));
        cleanup_records(records, num_files);
        close(fd);
        close(outfd);
        exit(16);
      }
      wrote += out;
    }

    /* Close the file descriptor */
    close(fd);
  }

  /* Cleanup */
  close(outfd);
  exit(0);
}
