#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>

void usage() {
  printf("tbfunpack: Crude .tbf file unpacker.\n");
  printf("usage: tbfunpack -l file.tbf (only lists the archive contents)\n");
  printf("usage: tbfunpack file.tbf [destdir]\n");
  exit(1);
}

int main(int argc, char **argv) {
  int fd, in, out, outfd, need, wrote;
  int num_files, i, offset, length, cur;
  char buffer[4096];
  char *filename;
  int doit;

  if (argc == 1) usage();

  /* Check -l argument */
  if (! strncmp(argv[1], "-l", 3)) {
    if (argc == 2) usage();
    argv++;
    argc--;
    doit = 0;
  }
  else doit = 1;

  /* Open file */
#ifdef _WIN32
  fd = open(argv[1], O_BINARY | O_RDONLY);
#else
  fd = open(argv[1], O_RDONLY);
#endif
  if (fd < 0) {
    fprintf(stderr, "Can't open tbf file '%s': %s\n", argv[1], strerror(errno));
    exit(1);
  }

  /* Change to output directory */
  if (argc > 2) {
    mkdir(argv[2], 0755); /* If it fails, tough */
    if (chdir(argv[2])) {
      fprintf(stderr, "Can't chdir '%s': %s\n", argv[2], strerror(errno));
      close(fd);
      exit(2);
    }
  }

  /* Read number of files */
  in = read(fd, &num_files, 4);
  if (in != 4) {
    fprintf(stderr, "Error reading number of files: %s\n", strerror(errno));
    close(fd);
    exit(3);
  }

  for (i = 0; i < num_files; i++) {
    /* Read filename length */
    in = read(fd, &length, 4);
    if (in != 4) {
      fprintf(stderr, "Error reading file #%d name length: %s\n", i, strerror(errno));
      close(fd);
      exit(4);
    }

    /* Allocate enough space */
    filename = (char *) malloc(length + 1);
    if (! filename) {
      fprintf(stderr, "Out of memory for file #%d name: %s\n", i, strerror(errno));
      close(fd);
      exit(5);
    }

    /* Read file */
    in = read(fd, filename, length);
    if (in != length) {
      fprintf(stderr, "Error reading file #%d name: %s\n", i, strerror(errno));
      free(filename);
      close(fd);
      exit(6);
    }
    filename[length] = '\0';

    /* Read offset */
    in = read(fd, &offset, 4);
    if (in != 4) {
      fprintf(stderr, "Error reading file #%d offset: %s\n", i, strerror(errno));
      free(filename);
      close(fd);
      exit(7);
    }

    /* Read length */
    in = read(fd, &length, 4);
    if (in != 4) {
      fprintf(stderr, "Error reading file #%d length: %s\n", i, strerror(errno));
      free(filename);
      close(fd);
      exit(8);
    }

    /* Write the file */
    if (doit) {
      printf("Extracting '%s'\n", filename);

      /* Remember file position */
      cur = lseek(fd, 0, SEEK_CUR);
      if (cur < 0) {
        fprintf(stderr, "Can't find descriptor position for file #%d ('%s'): %s\n", i, filename, strerror(errno));
        free(filename);
        close(fd);
        exit(9);
      }

      /* Seek to correct offset */
      if (lseek(fd, offset, SEEK_SET) != offset) {
        fprintf(stderr, "Can't seek to start of file file #%d ('%s'): %s\n", i, filename, strerror(errno));
        free(filename);
        close(fd);
        exit(10);
      }

      /* Open file for writing */
#ifdef _WIN32
      outfd = open(filename, O_BINARY | O_WRONLY | O_CREAT | O_TRUNC, 0644);
#else
      outfd = open(filename, O_WRONLY | O_CREAT | O_TRUNC, 0644);
#endif
      if (outfd < 0) {
        fprintf(stderr, "Can't open file #%d ('%s') for writing: %s\n", i, filename, strerror(errno));
        free(filename);
        close(fd);
        exit(11);
      }

      /* Write it! */
      wrote = 0;
      while (wrote < length) {
        /* What we need to write */
        need = length - wrote;
        if (need > sizeof(buffer)) need = sizeof(buffer);

        /* Read a bit */
        in = read(fd, &buffer, need);
        if (in != need) {
          fprintf(stderr, "Error reading from position %d of file '%s': %s\n", wrote, filename, strerror(errno));
          free(filename);
          close(fd);
          exit(12);
        }

        /* Write a bit */
        out = write(outfd, &buffer, in);
        if (out != in) {
          fprintf(stderr, "Error writing from position %d of file '%s': %s\n", i, filename, strerror(errno));
          free(filename);
          close(fd);
          exit(13);
        }
        wrote += out;
      }

      /* Close output file descriptor */
      close(outfd);

      /* Return to position */
      if (lseek(fd, cur, SEEK_SET) != cur) {
        fprintf(stderr, "Can't seek back from file #%d data: %s\n", i, strerror(errno));
        free(filename);
        close(fd);
        exit(14);
      }
    }
    else printf("File %d: '%s'\n", i + 1, filename);

    /* Cleanup */
    free(filename);
  }

  close(fd);
  exit(0);
}
